﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Windows.Forms;
using System.Management;
using UiO.TRAMP.Logger;
using UiO.TRAMP.Transport;
using System.Collections;
using System.Net.Sockets;
using System.Net;

namespace UiO.TRAMP.negotiation_and_data_transfer_module
{
    public partial class Form1 : Form
    {

        const string c_logfile_name = "negotiation_and_data_copy.log";
        const string c_interface_query = "Select DefaultIPGateway, IPAddress, Description, DNSDomain from Win32_NetworkAdapterConfiguration";
        const string c_DefaultIPGateway_property_name = "DefaultIPGateway";
        const string c_InterfaceDescription_property_name = "Description";
        const string c_IPAddress_property_name = "IPAddress";

        string tempFolder = Environment.GetEnvironmentVariable("TEMP");
        string username = Environment.GetEnvironmentVariable("USERNAME");
        string userdomain = Environment.GetEnvironmentVariable("USERDOMAIN");
        string computerName = Environment.GetEnvironmentVariable("COMPUTERNAME");

        Hashtable interfaceMap = new Hashtable();        

        string currentUser;
        string _selectedLocalIPAddress;
        LoggerClass Log;
        public NegotiationAndDataTransfer nadtConnectInstance = null;
        public NegotiationAndDataTransfer nadtBindInstance = null;
                
        public Form1()
        {
            InitializeComponent();
            InitializeRichTextBox();
            Log = new LoggerClass (c_logfile_name, tempFolder, richTextBox1);
            Log.OpenLogFile();            

            //connection = new TCPConnection(Log);

            currentUser = userdomain + @"\" + username;

            if (checkBox1.Checked == true)
            {
                Log.WriteToDebug = true;
                Log.WriteDebug("Form1(): WriteToDebug is sat to 'true' -> writing debug lines.");
            }
            else
            {
                Log.WriteDebug("Form1(): WriteToDebug is sat to 'false' -> ignoring debug lines.");
                Log.WriteToDebug = false;
            }

            if (checkBox2.Checked == true)
            {
                Log.WriteToLog = true;
                Log.WriteDebug("Form1(): WriteToDebug is sat to 'true' -> writing log lines.");
            }
            else
            {
                Log.WriteDebug("Form1(): WriteToDebug is sat to 'false' -> ignoring log lines.");
                Log.WriteToLog = false;
            }

            if (checkBox3.Checked == true)
            {
                Log.WriteToGUI = true;
                Log.WriteDebug("Form1(): WriteToGUI is sat to 'true' -> writing GUI lines.");
            }
            else
            {
                Log.WriteDebug("Form1(): WriteToGUI is sat to 'false' -> ignoring GUI lines.");
                Log.WriteToGUI = false;
            }

            InitializeInterfaceComboBox();            

            Log.WriteLog("Form1(): Logging a session for the user '" + currentUser + "' on computer '" + computerName.ToUpper() + "'.");

           
        }

        private void InitializeRichTextBox() {
            richTextBox1.Clear();
        }

        public String selectedLocalIPAddress {

            get {
                return _selectedLocalIPAddress;
            }

            set {
                _selectedLocalIPAddress = value;
            }
        }

        /// <summary>
        /// this function is used to initialize comboBox with the current network interfaces
        /// so that the user can choose which interface it wants to bind to
        /// </summary>
        private void InitializeInterfaceComboBox () {
            try {

                Log.WriteDebug("InitializeInterfaceComboBox(): Initializing ComboBox.");
                ManagementObjectSearcher mgmt = new ManagementObjectSearcher(c_interface_query);

                ManagementObjectCollection col = mgmt.Get();

                if (col.Count == 0)
                {
                    Log.WriteLogWarning("InitializeInterfaceComboBox(): Unable to find installed network interface. Aborting.");
                    return;
                }
                else {
                    Log.WriteDebug("InitializeInterfaceComboBox(): Found '" + col.Count + "' initialized network interfaces.");
                }

                int count = 0;

                foreach (ManagementObject adapter in col) {
                    
                    string descriptionMember = (string)adapter.GetPropertyValue(c_InterfaceDescription_property_name);                    

                    if (adapter.GetPropertyValue(c_DefaultIPGateway_property_name) != null)
                    {
                        count++;
                        Log.WriteLog("InitializeInterfaceComboBox(): Adding interface '" + descriptionMember + "' to the comboBox selecton.");
                        string[] ipaddressMember = (string[])adapter.GetPropertyValue(c_IPAddress_property_name);
                        comboBox1.Items.Add(descriptionMember.Trim());

                        interfaceMap.Add(descriptionMember.Trim(), ipaddressMember[0]);
                        Log.WriteLog("InitializeInterfaceComboBox(): Adding '" + descriptionMember + "' <-> '" + ipaddressMember[0] + "' to the interface hash table.");
                    }
                    else {
                        Log.WriteDebug("InitializeInterfaceComboBox(): Interface '" + descriptionMember + "' is not initialized.");
                    }
                }

                if (count == 0)
                {
                    Log.WriteLogWarning("InitializeInterfaceComboBox(): Unable to find initialized network interface. Aborting.");
                    return;
                }
                else {
                    Log.WriteDebugInformational("InitializeInterfaceComboBox(): Combobox initialized with total of '" + count + "' network interfaces.");
                }
            } catch (Exception exp) {
                Log.WriteLogWarning("InitializeInterfaceComboBox(): An error occured while initializing network interfaces. Error description: '" + exp.ToString() + "'.");                
            }
        }

        private void listBox1_SelectedIndexChanged(object sender, EventArgs e)
        {

        }
        /// <summary>
        /// The function that starts the sync bind process on the selected interface
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button1_Click(object sender, EventArgs e)
        {

            int localPort;

            if (comboBox1.SelectedItem == null) {
                Log.WriteLogWarning("button1_Click(): You have to select a local interface to bind to! Select one at the drop-down list by the name '" + label2.Text + "'.");
                return;
            }

            if (textBox2.Text.Length == 0) {
                Log.WriteLogWarning("button1_Click(): You have to select a local port to bind to! Write one in the text label with the name '" + textBox2.Text + "'.");
                return;
            } else {
                if (int.TryParse(textBox2.Text.ToString(), out localPort)) {
                    if (localPort < 1024) {
                        Log.WriteLogWarning("button1_Click(): Selected local port must be '> 1024'");
                        return;
                    } else {
                        Log.WriteDebug("button1_Click(): Selected local port number is '" + localPort + "'.");
                    }
                } else {
                    Log.WriteLogWarning("button1_Click(): Selected local port has to be a number: '1024 < local port < 65354'");
                    return;
                }
            }

            string key = comboBox1.SelectedItem.ToString();

            try
            {
                IPAddress localBindIP;

                if (IPAddress.TryParse(interfaceMap[key].ToString(), out localBindIP))
                {
                    selectedLocalIPAddress = interfaceMap[key].ToString();
                    Log.WriteLog("button1_Click(): retreived IPAddress object for ipaddress '" + interfaceMap[key].ToString() + "'.");
                }
                else
                {
                    Log.WriteLogWarning("button1_Click(): Unable to retreive ipaddress from string '" + interfaceMap[key].ToString() + "'. Aborting.");
                    return;
                }

                ThreadData tdata = new ThreadData(Log, localBindIP, localPort, ActionType.BIND, interfaceMap[key].ToString().Trim(), null);
                Thread workerThread = new Thread(new ParameterizedThreadStart(Run));
                
                Log.WriteLog("button1_Click(): Preparing the worker thread (TID: '" + workerThread.ManagedThreadId + "') for listening.");

                button1.Enabled = false;
                button2.Enabled = true;

                Log.WriteDebug("button1_click(): Disabling button1.");
                Log.WriteDebug("button1_click(): Enabling button2.");

                Log.WriteLog("button1_Click(): Starting the worker thread (TID: '" + workerThread.ManagedThreadId + "') for listening.");
                workerThread.Start(tdata);
           
            } catch (Exception exp) {
                Log.WriteLogWarning("button1_Click(): An error occured while trying to bind to the socket. Error code '" + exp.ToString() + "'.");
            }
            
        }

        private void button2_Click(object sender, EventArgs e)
        {
            string remoteHost;
            int remotePort;
            IPAddress remoteIP;
            IPHostEntry iphostInfo;

            if (textBox1.Text.Length == 0)
            {
                Log.WriteLogWarning("button2_Click(): you have to select a remote address or a host name to connect to!");
                return;
            }
            else
            {
                remoteHost = textBox1.Text.ToString();
                Log.WriteDebug("button2_Click(): Selected remote address or hostname is '" + remoteHost + "'.");
            }

            if (textBox3.Text.Length == 0)
            {
                Log.WriteLogWarning("button2_Click(): You have to select a remote port to connect to!");
                return;
            } else {
                if (int.TryParse(textBox3.Text.ToString(), out remotePort))
                {
                    if (remotePort < 1024) {
                        Log.WriteLogWarning("button2_Click(): Selected local port must be '> 1024'");
                        return;
                    } else {
                        Log.WriteDebug("button2_Click(): Selected local port number is '" + remotePort + "'.");
                    }
                } else {
                    Log.WriteLogWarning("button2_Click(): Selected local port has to be a number: '1024 < local port < 65354'");
                    return;
                }
            }

            if (IPAddress.TryParse(remoteHost, out remoteIP)) {
                Log.WriteLog("button2_Click(): retreived IPAddress object for ipaddress '" + remoteHost + "'.");
            } else {
                //try ti se if the name is valid DNS name
                try
                {
                    Log.WriteDebug("button2_Click(): Trying to find '" + remoteHost + "' in DNS");
                    iphostInfo = Dns.GetHostEntry(remoteHost);
                    remoteIP = iphostInfo.AddressList[0];
                }
                catch (Exception exp){
                    //throw (new Exception("Unable to resolve '" + remoteHost + "' in DNS. " + exp.ToString()));
                    Log.WriteLogWarning("button2_Click(): An error occured while trying to resolve hostname '" + remoteHost + "' in DNS. Error message: " + exp.ToString());
                    return;
                }
            }


            String localIP = selectedLocalIPAddress;
            ThreadData tdata = new ThreadData(Log, remoteIP, remotePort, ActionType.CONNECT, localIP, null);
            Thread workerThread = new Thread(new ParameterizedThreadStart(Run));

            Log.WriteLog("button2_Click(): Preparing the worker thread (TID: '" + workerThread.ManagedThreadId + "') for connection.");

            button2.Enabled = false;
            Log.WriteDebug("button2_click(): Disabling button2.");
            
            Log.WriteLog("button2_Click(): Starting the worker thread (TID: '" + workerThread.ManagedThreadId + "') for connection.");
            workerThread.Start(tdata);
        }

        public void Run(Object data) {

            ThreadData td = (ThreadData)data;

            Log.WriteDebug("Run(): Entering the main function of the worker thread and creating an instance of the 'NegotiationAndDataTransfer' class.");
            td.nadt = new NegotiationAndDataTransfer(td.Log, td.ipAddress, td.port, td.type, td.selectedIPAddress, progressBar2);
            
            if (td.type == ActionType.CONNECT) {
                nadtConnectInstance = td.nadt;
            } else {
                nadtBindInstance = td.nadt;
            }

            Log.WriteDebug("Run(): Current action is '" + Enum.GetName(typeof(ActionType), td.type) + "'.");

            td.nadt.Run();
        }

        /// <summary>
        /// A nice trick to make the scrollbar follow the new text
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void richTextBox1_TextChanged(object sender, EventArgs e)
        {
            richTextBox1.SelectionStart = richTextBox1.Text.Length;
            richTextBox1.ScrollToCaret();
        }

        private void Form1_Load(object sender, EventArgs e)
        {

        }

        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked == true) {
                Log.WriteToDebug = true;
                Log.WriteDebug("checkBox1_CheckedChanged(): WriteToDebug is sat to 'true' -> writing debug lines.");
            } else {
                Log.WriteDebug("checkBox1_CheckedChanged(): WriteToDebug is sat to 'false' -> ignoring debug lines.");
                Log.WriteToDebug = false;
            }
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e) {
            if (nadtConnectInstance != null)
                nadtConnectInstance.ReleaseResources();

            if (nadtBindInstance != null)
                nadtBindInstance.ReleaseResources();            

            Log.WriteToGUI = false;
            Log.WriteLog("************************* Closing Log *****************************");

            Application.Exit();
        }

        private void checkBox3_CheckedChanged(object sender, EventArgs e) {
            if (checkBox3.Checked == true) {
                Log.WriteToGUI = true;
                Log.WriteDebug("checkBox3_CheckedChanged(): WriteToGUI is sat to 'true' -> writing to GUI.");
            } else {
                Log.WriteDebug("checkBox1_CheckedChanged(): WriteToGUI is sat to 'false' -> skipping GUI update.");
                Log.WriteToGUI = false;
            }
        }

        private void checkBox2_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox2.Checked == true)
            {
                Log.WriteToLog = true;
                Log.WriteDebug("checkBox3_CheckedChanged(): WriteToLog is sat to 'true' -> writing log lines to file.");
            }
            else
            {
                Log.WriteDebug("checkBox1_CheckedChanged(): WriteToLog is sat to 'false' -> no updates to a log file.");
                Log.WriteToLog = false;
            }
        }     
    }

    public class ThreadData {

        public LoggerClass Log;
        public IPAddress ipAddress;
        public int port;
        public ActionType type;
        public String selectedIPAddress;
        public NegotiationAndDataTransfer nadt;

        public ThreadData(LoggerClass Log, IPAddress ipAddress, int port, ActionType type, String selectedIPAddress, NegotiationAndDataTransfer nadt) {
            this.Log = Log;
            this.ipAddress = ipAddress;
            this.port = port;
            this.type = type;
            this.selectedIPAddress = selectedIPAddress;
            this.nadt = nadt;
        }
    }
}
 