﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UiO.TRAMP.Logger;
using UiO.TRAMP.NegotiationProtocol;
using System.Net;
using System.Threading;
using System.Net.Sockets;
using System.Diagnostics;
using UiO.TRAMP.State;
using System.IO; //temporary, delete eventually
using UiO.TRAMP.negotiation_and_data_transfer_module;
using UiO.TRAMP.Transport;
using System.Xml;
using System.Xml.Schema;

namespace UiO.TRAMP.Negotiation

{
    class NegotiationClass
    {
        private const String c_rvns_namespace = "http://www.uio.no/XMLSchema/TRAMP/resourceVariables";

        public TCPConnection tcpConn; // will be used by rendezvous protocol (in the data transfer module)
        private LoggerClass Log;                
        private ActionType aType;

        public String selectedLocalIPAddress;
        public IPAddress ipAddress;

        public MigrationType migrationType;

        private int payloadLength = 0;
        private byte[] payload;

        Dictionary<XmlDocument, String> filePathToXMLDocMap = new Dictionary<XmlDocument, String>();

        private SIPPayloadState state;
        private NegotiationAlgorithmResult negotiationAlgorithmResult;

        private List<XmlDocument> appList = new List<XmlDocument>();

        string[] local_applications_file_path = {@"c:\temp\application_resource_needs.xml"};

        private Stopwatch sipSw = new Stopwatch();
        private TimeSpan sipTimeSpan;
        
        int port;

        //application used when the host is the offering host with both push and pull migration types
        const string c_application_resource_file_path = @"c:\temp\application_resource_needs.xml";
        //applications that exist on local host and that can be pulled from the remote host. It is therefore used only in pull migration type at the answering host 
        
        //schema file
        const string c_payload_schema_file_path = @"c:\temp\schema.xsd";
        //local resources
        const string c_local_resources_file_path = @"c:\temp\local_resources.xml";        

        NegotiationProtocolClass negotiationProtocol;

        public NegotiationClass(LoggerClass Log, String selectedLocalIPAddress, ActionType aType, IPAddress ipAddress, int port) {            
            
            this.Log = Log;
            this.selectedLocalIPAddress = selectedLocalIPAddress;
            this.aType = aType;
            this.port = port;
            this.ipAddress = ipAddress;            

            negotiationAlgorithmResult = NegotiationAlgorithmResult.OK; //default

            negotiationProtocol = new NegotiationProtocolClass(Log);
            tcpConn = negotiationProtocol.getTCPConnection;

            foreach (string pathToApp in local_applications_file_path) {
                XmlDocument xmlDoc = new XmlDocument();
                
                if(File.Exists(pathToApp)){
                    Log.WriteDebug("UiO.TRAMP.Negotiation.NegotiationClass(): Adding file '" + pathToApp + "' to the collection of local applications.");
                    xmlDoc.Load(pathToApp);
                    appList.Add(xmlDoc);
                    filePathToXMLDocMap.Add(xmlDoc, pathToApp);
                } else {
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.NegotiationClass(): Unable to find application file '" + pathToApp + "' on the local FS. Aborting.");
                    Abort();
                }
            }
        }

        private void Abort() {
            ReleaseResources();
            Thread.CurrentThread.Abort();
        }

        public Boolean Negotiate() {
            if (aType == ActionType.BIND) {
                NegotiateBind();
                ReceiveNegotiationRequest();
                RunNegotiationAlgorithm(out payload, out payloadLength);
                SendResponseToNegotiationRequest();
                ReceiveAckOnResponseToNegotiationRequest();
                //return true; // for now, in all other cases it shoud return whether the current host is source of upcomming data transfer or a destination
                return (migrationType == MigrationType.PUSH )? false : true; 
            } else {
                /*
                    Here I need to decide how will I signal the type of migration: push/pull?
                 * From GUI, defined by the user? It will not be a hudge change to the current situation since the c_application_resource_file_path will be replaced with 
                 * local resource file, and the migration type would be sat to the appropriate value
                 */
                NegotiateConnect();
                /*
                    This variable here must be sat externally, maybe in a GUI, such that we could choose the file that is going to be sent to the peer. For example:
                 * a) if migration is PUSH the file that is sent is local application resources needs, and the local host is the source host for the data transfer
                 * b) if migration is PULL, the file that is sent is the computer resources available localy, and the local host is the destination host for the data transfer
                 * 
                 * For now we say that the installation is always push
                 */
                //migrationType = MigrationType.PUSH;                
                //ValidateResourceVaraiblePresentationSystemInstanceFromFile(c_application_resource_file_path, c_payload_schema_file_path);
                //payload = ExtractPayloadFromXML(c_application_resource_file_path);
                migrationType = MigrationType.PULL;
                ValidateResourceVaraiblePresentationSystemInstanceFromFile(c_local_resources_file_path, c_payload_schema_file_path);
                payload = ExtractPayloadFromXML(c_local_resources_file_path);
                payloadLength = payload.Length;
                sipSw.Start();
                SendNegotiationRequest(ipAddress, selectedLocalIPAddress);
                ReceiveResposeAfterNegotiationRequest();
                AckResponseToNegotiationRequest();
                sipSw.Stop();
                if (Log.WriteToLog != true) {
                    Log.WriteToLog = true;
                    Log.WriteLog("UiO.TRAMP.Negotiation.Negotiate(): Writing the sip duration to the log file.");
                    sipTimeSpan = sipSw.Elapsed;
                    Log.WriteLog("UiO.TRAMP.Negotiation.Negotiate(): The sip protocol took '" + sipTimeSpan.TotalMilliseconds + "' ms.");
                    Log.WriteToLog = false;
                }
                return (migrationType == MigrationType.PUSH) ? true : false; 
            }
        }

        public void ReleaseResources() {
            Log.WriteLog("UiO.TRAMP.Negotiation.ReleaseResources(): Releasing network resources in the 'Negotiation' layer.");
            negotiationProtocol.ReleaseResources();
        }

        private void NegotiateBind()
        {
            Log.WriteLog("UiO.TRAMP.Negotiation.NegotiateBind(): Starting negotiation by awaiting negotiation request.");
            BindAndListenNegotiationProtocolSync(ipAddress, port);
        }

        private void NegotiateConnect()
        {
            Log.WriteLog("UiO.TRAMP.Negotiation.NegotiateConnect(): Starting negotiation by connecting to remote host.");
            ConnectNegotiationProtocolSync(ipAddress, port, selectedLocalIPAddress);
        }

        private void BindAndListenNegotiationProtocolSync(IPAddress ipAddress, int localPort)
        {
            Log.WriteDebug("UiO.TRAMP.Negotiation.BindAndListenNegotiationProtocolSync(): Binding to local endpoint '" + ipAddress.ToString() + ":" + localPort.ToString() + "' and waiting for negotiation request.");
            negotiationProtocol.BindAndListenNegotiationProtocolSync(ipAddress, localPort);
            /*
                This is where we end up after the remote TCP connections has been received. Here we need to wait for an INVITE from the counter-part
             */
        }

        private void ReceiveNegotiationRequest()
        {
            Log.WriteLog("UiO.TRAMP.Negotiation.ReceiveNegotiationRequest(): Entering negotiation protocol layer to listen for negotiation request.");
            negotiationProtocol.ReceiveNegotiationRequest();
            Log.WriteDebug("UiO.TRAMP.Negotiation.ReceiveNegotiationRequest(): Negotiation layer returned negotiation request. Running negotiation algorithm.");            
        }

        private void ConnectNegotiationProtocolSync(IPAddress ipAddress, int remotePort, String selectedLocalIPAddress)
        {
            Log.WriteDebug("UiO.TRAMP.Negotiation.ConnectNegotiationProtocolSync(): Starting connect from 'Negotiation' layer.");
            Log.WriteDebug("UiO.TRAMP.Negotiation.ConnectNegotiationProtocolSync(): Selected local address is '" + selectedLocalIPAddress + "'.");

            negotiationProtocol.ConnectNegotiationProtocolSync(ipAddress, remotePort);
            Log.WriteDebug("UiO.TRAMP.Negotiation.ConnectNegotiationProtocolSync(): Negotiation connection established.");
        }

        private void SendNegotiationRequest(IPAddress ipAddress, String selectedIPAddress) {
            if (payloadLength > 0) {
                Log.WriteLog("UiO.TRAMP.Negotiation.SendNegotiationRequest(): Sending negotiation request to the remote host.");
                negotiationProtocol.SendNegotiationRequest(ipAddress, selectedLocalIPAddress, payload, payloadLength);
                Log.WriteDebug("UiO.TRAMP.Negotiation.SendNegotiationRequest(): Returned from SendNegotiationRequest function.");
            } else {
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.SendNegotiationRequest(): An error occured while trying to send negotiation request. payload length is 0. Aborting");
                Abort();
            }
        }

        private void ValidateResourceVaraiblePresentationSystemInstanceFromFile(String file, String schemaFile) {            
            XmlReaderSettings settings = new XmlReaderSettings();
            XmlReader xmlReader;

            try
            {
                settings.ValidationType = ValidationType.Schema;
                settings.Schemas.Add(null, schemaFile);                
                settings.ValidationFlags |= XmlSchemaValidationFlags.ReportValidationWarnings;


                Log.WriteLog("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromFile(): Validating resource variable instance from file.");
                xmlReader = XmlReader.Create(file, settings); //this should be enough to just validate XML file against Schema. We do not need to parse the file at this point
            } catch (XmlSchemaValidationException exp) {
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromFile(): An error occured while validating resource variable instance file.");
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromFile(): " + exp.ToString());
                Abort();
            }
            catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromFile(): An error occured while loadig XML schema file. " + exp.ToString());
                Abort();
            }


            Log.WriteLogSuccess("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromFile(): Resource variable instance from file successfully validated.");
            settings = null;
            xmlReader = null;
        }

        private Boolean ValidateResourceVaraiblePresentationSystemInstanceFromStream(Stream xmlStream, String schemaFile) {
            XmlReaderSettings settings = new XmlReaderSettings();
            XmlReader xmlReader;

            try
            {
                settings.Schemas.Add(null, schemaFile);
                settings.ValidationType = ValidationType.Schema;
                settings.ValidationFlags |= XmlSchemaValidationFlags.ReportValidationWarnings;


                Log.WriteLog("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromStream(): Validating resource variable instance from stream.");
                xmlReader = XmlReader.Create(xmlStream, settings, ""); //this should be enough to just validate XML file against Schema. We do not need to parse the file at this point
            }
            catch (XmlSchemaValidationException exp)
            {
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromStream(): An error occured while validating resource variable instance file.");
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromStream(): " + exp.ToString());
                return false;
            }
            catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromStream(): " + exp.ToString());
                return false;
            }

            Log.WriteLogSuccess("UiO.TRAMP.Negotiation.ValidateResourceVaraiblePresentationSystemInstanceFromStream(): Resource variable instance from stream successfully validated.");
            settings = null;
            xmlReader = null;

            return true;
        }

        private void ReceiveResposeAfterNegotiationRequest () {
            Log.WriteLog("UiO.TRAMP.Negotiation.ReceiveResposeAfterNegotiationRequest(): Receiving response after negotiation request.");
            negotiationProtocol.ReceiveResposeAfterNegotiationRequest();
            Log.WriteLogSuccess("UiO.TRAMP.Negotiation.ReceiveResposeAfterNegotiationRequest(): Received response to negotiation request.");

            /*
             *  state variable, the one that carries the response payload to the "negotiation" layer will also have the information about the type of response 
             *  that has been received. This will be used to populate negotiationAlgorithm variable which in turn will equip  AckResponseToNegotiationRequest() to either continue to 
             *  data transfer (200 OK response) or abort (all other responses)
             *  
             * Other then that we are not taking any action based on the payload of the response. It is only the response code that matters
             */
        }

        private void AckResponseToNegotiationRequest() {
            Log.WriteLog("UiO.TRAMP.Negotiation.AckResponseToNegotiationRequest(): Acknowledging response to negotiation request.");
            negotiationProtocol.SendAckToNegotiationResponse();

            switch (negotiationProtocol.payloadState.rType) { 
                case SIPPayloadState.receivedResponseType.OK:
                    Log.WriteLogSuccess("UiO.TRAMP.Negotiation.AckResponseToNegotiationRequest(): Negotiation is finished. Moving on to the data transfer.");
                    break;
                case SIPPayloadState.receivedResponseType.NOTFOUND:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.AckResponseToNegotiationRequest(): Acknowledged negative negotiation response due to non-existant resources. Aborting.");
                    // Stop the negotiation process
                    Abort();
                    break;
                case SIPPayloadState.receivedResponseType.NOTACCEPTABLEHERE:
                    // Stop the negotiation process
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.AckResponseToNegotiationRequest(): Acknowledged negative negotiation response due to insufficient resources. Aborting.");
                    Abort();
                    break;
                case SIPPayloadState.receivedResponseType.BADREQUEST:
                    // Stop the negotiation process
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.AckResponseToNegotiationRequest(): Acknowledged negative negotiation response due to bad request. Aborting.");
                    Abort();
                    break;
                default:
                    break;
            }
        }

        private void SendResponseToNegotiationRequest() {
            switch (negotiationAlgorithmResult)
            { 
                case NegotiationAlgorithmResult.OK:
                    Log.WriteLogSuccess("UiO.TRAMP.Negotiation.SendResponseToNegotiationRequest(): Negotiation is possible. Returning positive response.");
                    negotiationProtocol.SendPositiveNegotiationResponse(payload, payloadLength);
                    Log.WriteDebug("UiO.TRAMP.Negotiation.SendResponseToNegotiationRequest(): Positive response sent.");
                    break;
                case NegotiationAlgorithmResult.NOTACCEPTABLEHERE:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.SendResponseToNegotiationRequest(): Negotiation is not possible due to insufficient resources. Returning negative response.");                    
                    negotiationProtocol.SendNegativeNegotiationResponse(payload, payloadLength);
                    Log.WriteDebug("UiO.TRAMP.negotiation.SendResponseToNegotiationRequest(): Negative response sent.");
                    break;
                case NegotiationAlgorithmResult.BADREQUEST:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.SendResponseToNegotiationRequest(): Negotiation is not possible due to malformed request. Returning negative response.");
                    negotiationProtocol.SendNegativeNegotiationResponseBadRequest(payload, payloadLength);
                    Log.WriteDebug("UiO.TRAMP.negotiation.SendResponseToNegotiationRequest(): Negative response sent.");
                    break;
                case NegotiationAlgorithmResult.NOTFOUND:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.SendResponseToNegotiationRequest(): Negotiation is not possible since wanted application is not found. Returning negative response.");
                    negotiationProtocol.SendNegativeNegotiationResponseNotFound(payload, payloadLength);
                    Log.WriteDebug("UiO.TRAMP.negotiation.SendResponseToNegotiationRequest(): Negative response sent.");
                    break;
                default:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.SendResponseToNegotiationRequest(): Unknown result from negotiation algorthm. Aborting.");
                    Abort();
                    break;
            }
        }

        private void RunNegotiationAlgorithm(out byte [] payload, out int payloadLength) {
            /*
             *  after the negotiation algorithm has finished running it will populate the payload and payload length with the answer and also the 
             *  negotiationAlgorithmResult variable which will then decide which answer will be sent to the peer
             */

            XmlDocument rcvResources = new XmlDocument();
            XmlDocument locResources = new XmlDocument();

            /*
                These two variables represent collections of resources for both received payload and local files
             * The way they are initialized is dependant on the migration type, that is: for the push migration 
             * the received payload is compared against local hw resources while for the pull migration the local app state is compared with 
             * received payload that represents the hw state on the offering host
             */
            XmlNodeList appResList;
            XmlNodeList resResList;

            //here we will put the answer of the negotiation algorithm
            XmlDocument answerXMLDocument = new XmlDocument();
            
            //creating the root element of the answer file since we return at least that 
            XmlElement rvs = answerXMLDocument.CreateElement("resourceVariables");
            answerXMLDocument.AppendChild(rvs);

            negotiationAlgorithmResult = NegotiationAlgorithmResult.OK; // we begin by using 'optimistic' approach ... ;-)
            Log.WriteLog("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Running negotiation algorithm.");

            /*
                Tasks:
             * 1) Load
             * 2) Validate received xml file
             * 3) Extract migration type
             * 4) validate local resources file
             * 5) compare
             * 6) return
             */

            // extracting payload state from the negotiation layer
            state = negotiationProtocol.payloadState;

            byte[] rcvPayload = state.content;
            int rcvPayloadLength = state.contentLength;

            MemoryStream rcvStream = new MemoryStream();
            rcvStream.Write(rcvPayload, 0, rcvPayloadLength);            

            //validating the received payload against schema - if something is wrong the processing stops here
            if (!(ValidateResourceVaraiblePresentationSystemInstanceFromStream(rcvStream, c_payload_schema_file_path))) {
                negotiationAlgorithmResult = NegotiationAlgorithmResult.BADREQUEST;
                payload = rcvPayload; //returning the payload just the way it was, since the request was malformed
                payloadLength = rcvPayloadLength;
                return;
            }

            //extracting migration type
            //appResources.Load(rcvStream);
            rcvResources.LoadXml(Encoding.UTF8.GetString(rcvPayload));

            Log.WriteLogInformational(Encoding.UTF8.GetString(rcvPayload));

            XmlNode directionNode = rcvResources.SelectSingleNode("//direction");
            XmlNode appNameNode = rcvResources.SelectSingleNode("//appName");

            Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Read application name '" + appNameNode.InnerText + "'.");

            Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Reading migration type.");
            if (directionNode.InnerText.Equals("push", StringComparison.OrdinalIgnoreCase)) {
                Log.WriteLog("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Setting migration type to '" + Enum.GetName(typeof(MigrationType), 1) + "'.");
                migrationType = MigrationType.PUSH;                         
                //in the case of the push migration type the local file of importance is the local resource state
                ValidateResourceVaraiblePresentationSystemInstanceFromFile(c_local_resources_file_path, c_payload_schema_file_path); // consider here a true/false in order to return some error code if validation fails
                locResources.Load(c_local_resources_file_path);

                appResList = rcvResources.SelectNodes("//resourceVariable");
                resResList = locResources.SelectNodes("//resourceVariable");
            } else {
                Log.WriteLog("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Setting migration type to '" + Enum.GetName(typeof(MigrationType), 2) + "'.");
                migrationType = MigrationType.PULL;

                String localAppFilePath; // in the case of the pull negotiation type it is the local application's resources that need to be validated

                Log.WriteLog("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Checking whether wanted application exists on the local host.");
                if (IsAppMigrateble(rcvResources, out localAppFilePath)) {
                    Log.WriteLogSuccess("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Found wanted aplication on the local host.");
                    ValidateResourceVaraiblePresentationSystemInstanceFromFile(localAppFilePath, c_payload_schema_file_path);
                    locResources.Load(localAppFilePath);

                    appResList = locResources.SelectNodes("//resourceVariable");
                    resResList = rcvResources.SelectNodes("//resourceVariable");
                } else {
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Unable to find wanted application on local host. Sending 'NOT FOUND' response.");
                    negotiationAlgorithmResult = NegotiationAlgorithmResult.NOTFOUND;
                    payload = rcvPayload; //returning the payload just the way it was, since the request was malformed
                    payloadLength = rcvPayloadLength;
                    return;
                }
            }

            //Adding 'direction' and 'appName' elements to the answerXML file
            XmlElement dir = answerXMLDocument.CreateElement("direction");
            dir.InnerText = directionNode.InnerText;

            XmlElement appName = answerXMLDocument.CreateElement("appName");
            appName.InnerText = appNameNode.InnerText;
            
            rvs.AppendChild(dir);
            rvs.AppendChild(appName);

            Boolean foundRVNameMatch = false;
            Boolean foundResourceMatch = false;                        

            Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Found total of '" + appResList.Count + "' resource variable nodes in received XML file.");
            Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Found total of '" + resResList.Count + "' resource variable nodes in local XML file.");
            // inspecting the payload, comparing it with the local resources and creating an answer
            foreach (XmlNode appNode in appResList) {
                String rcvRVName = appNode.SelectSingleNode("name").InnerText;                
                foreach (XmlNode localNode in resResList) {                    
                    String locRVName = localNode.SelectSingleNode("name").InnerText;
                    Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Comparing (RCV) '" + rcvRVName + "' <-> (LOC)'" + locRVName + "' resource variables.");
                    if (rcvRVName.Equals(locRVName, StringComparison.OrdinalIgnoreCase)) {
                        Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Found a resource variable (name) match.");
                        foundRVNameMatch = true;

                        //Create resourceVariable node and its name child node
                        XmlElement rv = answerXMLDocument.CreateElement("resourceVariable");
                        XmlElement rvName = answerXMLDocument.CreateElement("name");
                        
                        rvName.InnerText = rcvRVName;
                        rv.AppendChild(rvName);

                        XmlNodeList appResourcesList = appNode.SelectNodes("resource");
                        XmlNodeList locResourceList = localNode.SelectNodes("resource");

                        Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Found total of '" + appResourcesList.Count + "' resource variable property nodes in received XML file.");
                        Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Found total of '" + locResourceList.Count + "' resource variable property nodes in local XML file.");

                        //now it is time to traverse 'resource' XML elements
                        foreach (XmlNode appNodeResources in appResourcesList) {
                            String appResourceName = appNodeResources.SelectSingleNode("pn").InnerText;
                            foreach (XmlNode localNodeResources in locResourceList) {
                                String locResourceName = localNodeResources.SelectSingleNode("pn").InnerText;
                                Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Comparing (appResourceName) '" + appResourceName + "' <-> (locResourceName)'" + locResourceName + "' resource variable.");
                                if (appResourceName.Equals(locResourceName, StringComparison.OrdinalIgnoreCase)) {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Found a resource match.");
                                    foundResourceMatch = true;

                                    XmlNode importNode;

                                    ResourceMatchType resmatchTypeResult = ResourceMatch(appNodeResources, localNodeResources);

                                    switch (resmatchTypeResult)
                                    {
                                        case ResourceMatchType.MATCH:
                                            /*
                                                If we have a match, all we need to do is to add the received resource and send it back to the offering host 
                                             * In this way, if the final response is positive, we will signal the offering host that this resource is supported
                                             */
                                            importNode = rv.OwnerDocument.ImportNode(appNodeResources,true);
                                            //rv.AppendChild(appNodeResources);
                                            rv.AppendChild(importNode);
                                            break; 
                                        case ResourceMatchType.NOMATCH:
                                            /*
                                                If it is however no-match, we are sending the current local state of the wanted resource so that the offering host has an idea 
                                             * about what resources are available at the answering host
                                             */
                                            importNode = rv.OwnerDocument.ImportNode(localNodeResources, true);
                                            //rv.AppendChild(localNodeResources);
                                            rv.AppendChild(importNode);
                                            negotiationAlgorithmResult = NegotiationAlgorithmResult.NOTACCEPTABLEHERE;
                                            break; 
                                        case ResourceMatchType.SUPPORTED:
                                            /*
                                                The situation here is the following: match was not found, but since the resource was not 'required' we could say that it was accepted
                                             *  Difference relative to the previous case is that here we return positive answer, such that the offering host has some idea about the 
                                             *  current resource state on the answering host
                                             */
                                            importNode = rv.OwnerDocument.ImportNode(localNodeResources, true);
                                            //rv.AppendChild(localNodeResources);
                                            rv.AppendChild(importNode);
                                            break;
                                        default:
                                            //not possible
                                            break;
                                    }

                                    break; // inner foreach
                                }                                
                            }

                            if (!foundResourceMatch) {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): No matching resource was found for '" + appResourceName + "' for the resource variable '" + rcvRVName + "'.");
                                //foundResourceMatch = false; //reset
                                /*
                                    Allready here we can decide tha this negotiation is not going to be success since an entire resource variable is not found
                                 *  We continue however our search since we need to return an answer to the offering host anyway
                                 */
                                negotiationAlgorithmResult = NegotiationAlgorithmResult.NOTACCEPTABLEHERE;
                            }
                        }

                        rvs.AppendChild(rv);
                        break; // inner foreach
                    } 
                }

                if (!foundRVNameMatch) {
                    Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): No match was found for '" + rcvRVName + "' resource variable.");
                    //foundRVNameMatch = false; //reset
                    /*
                        Allready here we can decide tha this negotiation is not going to be success since an entire resource variable is not found
                     *  We continue however our search since we need to return an answer to the offering host anyway
                     */
                    negotiationAlgorithmResult = NegotiationAlgorithmResult.NOTACCEPTABLEHERE;
                }                
            }

            Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): The calculated result is: '" + Enum.GetName(typeof(NegotiationAlgorithmResult), negotiationAlgorithmResult) + "'.");

            //just for debugging purposes
            Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): The constructed answer file is: \r\n\r\n" + rvs.OuterXml);

            payload = Encoding.UTF8.GetBytes(rvs.OuterXml);            
            payloadLength = payload.Length;

            Log.WriteDebug("UiO.TRAMP.Negotiation.RunNegotiationAlgorithm(): Constructed response payload of '" + payloadLength + "' B in length.");            
        }

        private ResourceMatchType ResourceMatch(XmlNode rcvApp, XmlNode locApp) {
            String appReq;
            String appValue;
            String appUnit;
            String appCo;
            String appRangeMinimum = "";
            String appRangeMaximum = "";

            String locReq;
            String locValue;
            String locUnit;
            String locCo;
            String locRangeMinimum = "";
            String locRangeMaximum = "";

            try
            {
                //parsing received app resource needs 
                appReq = rcvApp.SelectSingleNode("required").InnerText;
                appValue = rcvApp.SelectSingleNode("value").InnerText;
                appUnit = rcvApp.SelectSingleNode("unit").InnerText;
                appCo = rcvApp.SelectSingleNode("co").InnerText;

                if (rcvApp.SelectSingleNode("minimum") != null)
                    appRangeMinimum = rcvApp.SelectSingleNode("minimum").InnerText;
                if (rcvApp.SelectSingleNode("maximum") != null)
                    appRangeMaximum = rcvApp.SelectSingleNode("maximum").InnerText;

                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking received resources: (req) " + appReq + ", (value) " + appValue + ", (unit) " + appUnit + ", (co) " + appCo);

                if (appCo == "range") {
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking received resources: (minumun) " + appRangeMinimum + ", (maximum) " + appRangeMaximum);
                }


                //parsing local resources 
                locReq = locApp.SelectSingleNode("required").InnerText;
                locValue = locApp.SelectSingleNode("value").InnerText;
                locUnit = locApp.SelectSingleNode("unit").InnerText;
                locCo = locApp.SelectSingleNode("co").InnerText;
                if (locApp.SelectSingleNode("minimum") != null)
                    locRangeMinimum = locApp.SelectSingleNode("minimum").InnerText;
                if (locApp.SelectSingleNode("maximum") != null)
                    locRangeMaximum = locApp.SelectSingleNode("maximum").InnerText;

                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking local resources: (req) " + locReq + ", (value) " + locValue + ", (unit) " + locUnit + ", (co) " + locCo);

                if (locCo == "range") {
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking local resources: (minumun) " + locRangeMinimum + ", (maximum) " + locRangeMaximum);
                }
            } catch (Exception exp) {
                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources " + exp.ToString());
                return ResourceMatchType.NOMATCH;
            }

            /*
                The basic assumption here is that the received resource is compared against the value of local resource 
             * That is we will not take into consideration the comparisson opperator from the local resource but 
             * rather assume that the value represents the appropriate estimate of the local resource
             */

            int aValue;
            int lValue; 

            switch (appCo) {
                case "eq":
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking resources for operator 'eq'.");
                    if (int.TryParse(appValue, out aValue)) {
                        if (int.TryParse(locValue, out lValue)) {
                            /*
                                We will assume that units always match and for now will not make any attempt on converting between 
                             * different unit types in order to have this more precise
                             */
                            if ((aValue == lValue) && (appUnit.Equals(locUnit, StringComparison.OrdinalIgnoreCase))) {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and units match for 'eq' operator. Returning 'MATCH' resourcematchType.");
                                return ResourceMatchType.MATCH;
                            } else {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and/or units don't match for 'eq' operator. Returning 'NOMATCH' resourcematchType.");
                                if (appReq == "true") {
                                    return ResourceMatchType.NOMATCH;
                                } else {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Required element is sat to 'false'. Returning 'SUPPORTED' resourcematchType.");
                                    return ResourceMatchType.SUPPORTED;
                                }
                            }
                        } else {
                            // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                            Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                            Abort();
                        }
                    } else {
                        // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                        Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                        Abort();
                    }
                    break;
                case "ne":
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking resources for operator 'ne'.");
                    if (int.TryParse(appValue, out aValue)) {
                        if (int.TryParse(locValue, out lValue)) {
                            /*
                                We will assune that units always match and for now will not make any attempt on converting between 
                             * different unit types in order to have this more precise
                             */
                            if ((aValue != lValue) && (appUnit.Equals(locUnit, StringComparison.OrdinalIgnoreCase))) {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and units match for 'ne' operator. Returning 'MATCH' resourcematchType.");
                                return ResourceMatchType.MATCH;
                            } else {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and/or units don't match for 'ne' operator. Returning 'NOMATCH' resourcematchType.");
                                if (appReq == "true") {
                                    return ResourceMatchType.NOMATCH;
                                } else {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Required element is sat to 'false'. Returning 'SUPPORTED' resourcematchType.");
                                    return ResourceMatchType.SUPPORTED;
                                }
                            }
                        } else {
                            // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                            Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                            Abort();
                        }
                    } else {
                        // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                        Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                        Abort();
                    }
                    break;
                case "gt":
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking resources for operator 'gt'.");
                    if (int.TryParse(appValue, out aValue)) {
                        if (int.TryParse(locValue, out lValue)) {
                            /*
                                We will assune that units always match and for now will not make any attempt on converting between 
                             * different unit types in order to have this more precise
                             */
                            if ((lValue > aValue) && (appUnit.Equals(locUnit, StringComparison.OrdinalIgnoreCase))) {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and units match for 'gt' operator. Returning 'MATCH' resourcematchType.");
                                return ResourceMatchType.MATCH;
                            } else {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and/or units don't match for 'gt' operator. Returning 'NOMATCH' resourcematchType.");
                                if (appReq == "true") {
                                    return ResourceMatchType.NOMATCH;
                                } else {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Required element is sat to 'false'. Returning 'SUPPORTED' resourcematchType.");
                                    return ResourceMatchType.SUPPORTED;
                                }
                            }
                        } else {
                            // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                            Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                            Abort();
                        }
                    } else {
                        // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                        Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                        Abort();
                    }
                    break;
                case "lt":
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking resources for operator 'lt'.");
                    if (int.TryParse(appValue, out aValue)) {
                        if (int.TryParse(locValue, out lValue)) {
                            /*
                                We will assune that units always match and for now will not make any attempt on converting between 
                             * different unit types in order to have this more precise
                             */
                            if ((lValue < aValue) && (appUnit.Equals(locUnit, StringComparison.OrdinalIgnoreCase))) {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and units match for 'lt' operator. Returning 'MATCH' resourcematchType.");
                                return ResourceMatchType.MATCH;
                            } else {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and/or units don't match for 'lt' operator. Returning 'NOMATCH' resourcematchType.");
                                if (appReq == "true") {
                                    return ResourceMatchType.NOMATCH;
                                } else {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Required element is sat to 'false'. Returning 'SUPPORTED' resourcematchType.");
                                    return ResourceMatchType.SUPPORTED;
                                }
                            }
                        } else {
                            // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                            Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                            Abort();
                        }
                    } else {
                        // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                        Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                        Abort();
                    }
                    break;
                case "le":
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking resources for operator 'le'.");
                    if (int.TryParse(appValue, out aValue)) {
                        if (int.TryParse(locValue, out lValue)) {
                            /*
                                We will assune that units always match and for now will not make any attempt on converting between 
                             * different unit types in order to have this more precise
                             */
                            if ((lValue <= aValue) && (appUnit.Equals(locUnit, StringComparison.OrdinalIgnoreCase))) {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and units match for 'le' operator. Returning 'MATCH' resourcematchType.");
                                return ResourceMatchType.MATCH;
                            } else {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and/or units don't match for 'le' operator. Returning 'NOMATCH' resourcematchType.");
                                if (appReq == "true") {
                                    return ResourceMatchType.NOMATCH;
                                } else {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Required element is sat to 'false'. Returning 'SUPPORTED' resourcematchType.");
                                    return ResourceMatchType.SUPPORTED;
                                }
                            }
                        } else {
                            // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                            Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                            Abort();
                        }
                    } else {
                        // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                        Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                        Abort();
                    }
                    break;
                case "ge":
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking resources for operator 'ge'.");
                    if (int.TryParse(appValue, out aValue)) {
                        if (int.TryParse(locValue, out lValue)) {
                            /*
                                We will assune that units always match and for now will not make any attempt on converting between 
                             * different unit types in order to have this more precise
                             */
                            if ((lValue >= aValue) && (appUnit.Equals(locUnit, StringComparison.OrdinalIgnoreCase))) {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and units match for 'ge' operator. Returning 'MATCH' resourcematchType.");
                                return ResourceMatchType.MATCH;
                            } else {
                                Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and/or units don't match for 'ge' operator. Returning 'NOMATCH' resourcematchType.");
                                if (appReq == "true") {
                                    return ResourceMatchType.NOMATCH;
                                } else {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Required element is sat to 'false'. Returning 'SUPPORTED' resourcematchType.");
                                    return ResourceMatchType.SUPPORTED;
                                }
                            }
                        } else {
                            // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                            Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                            Abort();
                        }
                    } else {
                        // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                        Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                        Abort();
                    }
                    break;
                case "range":
                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Checking resources for operator 'range'.");
                    if (int.TryParse(appValue, out aValue)) {
                        if (int.TryParse(locValue, out lValue)) {
                            /*
                                We will assune that units always match and for now will not make any attempt on converting between 
                             * different unit types in order to have this more precise
                             */
                            int min, max;

                            if ((int.TryParse(appRangeMinimum, out min)) && (int.TryParse(appRangeMaximum, out max)))
                            {
                                if ((min <= lValue) && (lValue <= max) && (appUnit.Equals(locUnit, StringComparison.OrdinalIgnoreCase))) {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and units match for 'range' operator. Returning 'MATCH' resourcematchType.");
                                    return ResourceMatchType.MATCH;
                                } else {
                                    Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Values and/or units don't match for 'range' operator. Returning 'NOMATCH' resourcematchType.");
                                    if (appReq == "true")
                                    {
                                        return ResourceMatchType.NOMATCH;
                                    }
                                    else
                                    {
                                        Log.WriteDebug("UiO.TRAMP.Negotiation.ResourceMatch(): Required element is sat to 'false'. Returning 'SUPPORTED' resourcematchType.");
                                        return ResourceMatchType.SUPPORTED;
                                    }
                                }
                            } else {
                                Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                                Abort();
                            }
                        } else {
                            // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                            Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                            Abort();
                        }
                    } else {
                        // interneal server error, unlikely to happen but since we do not return 5xx responses here we will just abort thread
                        Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): An error occured while parsing resources");
                        Abort();
                    }
                    break;
                default:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.ResourceMatch(): Unsupported comparison operator '" + appCo + "'. Aborting.");
                    Abort();
                    break;
            }
            return ResourceMatchType.NOMATCH;
        }

        private Boolean IsAppMigrateble(XmlDocument xmlDoc, out String localAppFilePath) {
            XmlNode wantedApp = xmlDoc.SelectSingleNode("//appName");

            Log.WriteDebug("UiO.TRAMP.Negotiation.IsAppMigrateble(): Searching for '" + wantedApp.InnerText + "' application on local host.");
            Log.WriteDebug("UiO.TRAMP.Negotiation.IsAppMigrateble(): Searching a list with total of '" + appList.Count + "' local applications.");

            foreach (XmlDocument appXMLDoc in appList) {
                XmlNode localApp = appXMLDoc.SelectSingleNode("//appName");
                Log.WriteDebug("UiO.TRAMP.Negotiation.IsAppMigrateble(): Inspecting '" + localApp.InnerText + "' application on local host.");
                if (wantedApp.InnerText.Equals(localApp.InnerText, StringComparison.OrdinalIgnoreCase)) {
                    Log.WriteDebug("UiO.TRAMP.Negotiation.IsAppMigrateble(): Found a match on local host: '" + wantedApp.InnerText + "<->" + localApp.InnerText + "'." );
                    localAppFilePath = filePathToXMLDocMap[appXMLDoc];
                    return true;
                }
            }

            localAppFilePath = ""; 
            return false;
        }

        public void ReceiveAckOnResponseToNegotiationRequest() {
            negotiationProtocol.ReceiveAckOnResponseToNegotiationRequest();

            switch (negotiationAlgorithmResult)
            {
                case NegotiationAlgorithmResult.OK:
                    Log.WriteLogSuccess("UiO.TRAMP.Negotiation.ReceiveAckOnResponseToNegotiationRequest(): Received ack on positive response to negotiation request.");
                    // continuing with data transfer
                    break;
                case NegotiationAlgorithmResult.NOTFOUND:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.ReceiveAckOnResponseToNegotiationRequest(): Received ack on negative response because wanted application is not found. Stopping negotiation process.");
                    // Stop the negotiation process
                    Abort();
                    break;
                case NegotiationAlgorithmResult.BADREQUEST:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.ReceiveAckOnResponseToNegotiationRequest(): Received ack on negative response due to malformed request. Stopping negotiation process.");
                    // Stop negotiation process
                    Abort();
                    break;
                case NegotiationAlgorithmResult.NOTACCEPTABLEHERE:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.ReceiveAckOnResponseToNegotiationRequest(): Received ack on negative response due to insufficient resources. Stopping negotiation.");
                    // Stop negotiation process
                    Abort();
                    break;
                default:
                    Log.WriteLogWarning("UiO.TRAMP.Negotiation.ReceiveAckOnResponseToNegotiationRequest(): Unknown result from negotiation algorthm. Aborting.");
                    Abort();
                    break;
            }
        }

        private byte[] ExtractPayloadFromXML(String pathToXML) {

            int payloadLength;
            byte[] payload;

            Log.WriteLog("UiO.TRAMP.Negotiation.ExtractPayloadFromXML(): Extracting negotiation payload.");

            try {
                if (File.Exists(pathToXML)) {

                    Log.WriteDebug("UiO.TRAMP.Negotiation.ExtractPayloadFromXML(): Found negotiation payload instance file: '" + pathToXML + "'.");

                    using (FileStream fs = new FileStream(pathToXML, FileMode.Open, FileAccess.Read)) {
                        
                        payloadLength = (int)fs.Length;
                        payload = new byte[payloadLength];

                        Log.WriteDebug("UiO.TRAMP.Negotiation.ExtractPayloadFromXML(): Payload length is: '" + payloadLength + "'.");

                        int bytesRead = fs.Read(payload, 0, payloadLength);
                        Log.WriteDebug("UiO.TRAMP.Negotiation.ExtractPayloadFromXML(): Extracted '" + payloadLength + "' B from file '" + pathToXML + "'.");

                        return payload;
                    }
                } else {
                    Log.WriteDebug("UiO.TRAMP.negotiation.ExtractPayloadFromXML(): Unable to find file: '" + pathToXML + "'. Aborting.");
                    Abort();
                }
            } catch (Exception exp) {
                Log.WriteDebug("UiO.TRAMP.negotiation.ExtractPayloadFromXML(): An error occured while extracting answer file. " + exp.ToString());
                Abort();
            }

            return new byte[1]; // VS asks something to be returned here 
        }

    }

    public enum MigrationType { 
        PUSH = 1,
        PULL
    }

    public enum ResourceMatchType { 
        MATCH,
        NOMATCH,
        SUPPORTED
    }
}
