﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Net.Sockets;
using System.Net;
using System.Diagnostics;
using UiO.TRAMP.Negotiation;
using UiO.TRAMP.Logger;
using UiO.TRAMP.SIP;
using UiO.TRAMP.DataTransfer;

namespace UiO.TRAMP.negotiation_and_data_transfer_module
{
    public class NegotiationAndDataTransfer
    {
        private LoggerClass Log;
        private IPAddress ipAddress;
        private int port;
        private Boolean source;
        private NegotiationClass nc;
        private DataTransferClass dt;
        //private ActionType type;
        private String selectedLocalIPAddress;

        private Stopwatch negotiationSw = new Stopwatch();
        private TimeSpan negotiationTimeSpan; 

        public NegotiationAndDataTransfer(LoggerClass Log, IPAddress ipAddress, int port, ActionType aType, String selectedLocalIPAddress, Object gui)
        {
            this.Log = Log;
            this.ipAddress = ipAddress;
            this.port = port;
            //this.type = type;
            this.selectedLocalIPAddress = selectedLocalIPAddress;
            nc = new NegotiationClass(Log, selectedLocalIPAddress, aType, ipAddress, port);
            dt = new DataTransferClass(Log, selectedLocalIPAddress, gui);
        }

        public void Run() {
            //This function runs the whole show when this side is receiving the first negotiation request
            Log.WriteLog("UiO.TRAMP.Run(): Starting negotiation by entering the 'Negotiation' layer.");
            negotiationSw.Start();
            source = nc.Negotiate();
            negotiationSw.Stop();
            // Write the elapsed negotiation time to the log file
            if (Log.WriteToLog != true) {
                Log.WriteToLog = true;
                Log.WriteLog("UiO.TRAMP.Run(): Writing the negotiation duration to the log file.");
                negotiationTimeSpan = negotiationSw.Elapsed;
                Log.WriteLog("UiO.TRAMP.Run(): The negotiation protocol took '" + negotiationTimeSpan.TotalMilliseconds + "' ms.");
                Log.WriteToLog = false;
            }
            dt.TransferData(source, nc.tcpConn, nc.ipAddress);
            //ReleaseResources();
        }

        public void ReleaseResources() { 
            /*
                This function releases all the dialogs, states and sockets in both negotiation and data transfer classes
             */
            nc.ReleaseResources();
            //dt.ReleaseResources();
        }

        private void Abort() {
            ReleaseResources();
            Thread.CurrentThread.Abort();
        }

    }
}
